;; BF IR Emitter
;; Jon Simons (simonsj at ccs dot neu dot edu)
;; August, 2007
;; -------------------------------------------------------------------------
;; This module emits the BF Intermediate Representation.
;;
;;   The emitted code is just a pseudo-S expression that can be used
;;   to visualize the BF IR.
;;
;;   Here is a "key" for interpreting how the different IR constructs
;;   are output:
;;
;;     P is a number representing a pointer location.  For LOOPs and MULs,
;;     this is the pointer location at the beginning of that expression.
;;     For all other operations, this is the pointer to the memory location
;;     being operated on.
;;
;;     N is a number representing how many times the given operation is
;;     performed.
;;
;;     (INC N P)
;;     (DEC N P)
;;     (LT  N)
;;     (RT  N)
;;     (OUT P)
;;     (IN  P)
;;     (LOOP P <EXPRS> ...)
;;     (MUL  P <EXPRS> ...)
;; -------------------------------------------------------------------------
(module bf-emit-ir mzscheme
  
  (provide emit)
  
   ;; For AST structs
  (require "../bf-ast-transform.scm")
  ;; Emit Utils
  (require "bf-emit-util.scm")
  
  (require (lib "list.ss"))
  
  ;; emit : (list of EXPRs) -> string
  (define (emit AST)
    (apply string-append (reverse (emit-helper AST () 0))))
  
  (define (emit-helper AST acc indent)
    (if (null? AST)
        acc
        (let* ((expr  (car AST))
               (type  (EXPR-type expr))
               (N     (EXPR-N expr))
               (ptr   (EXPR-ptr expr))
               (exprs (EXPR-exprs expr))
               (str  (case type
                       ((inc) (pad (format "(INC ~v ~v)~%" N ptr) indent))
                       ((dec) (pad (format "(DEC ~v ~v)~%" N ptr) indent))
                       ((rt)  (pad (format "(RT ~v)~%" N) indent))
                       ((lt)  (pad (format "(LT ~v)~%" N) indent))
                       ((out) (pad (format "(OUT ~v)~%" ptr) indent))
                       ((in)  (pad (format "(IN ~v)~%" ptr) indent))
                       ((loop) (string-append
                                (pad (format "(LOOP ~v~%" ptr) indent)
                                (apply string-append
                                       (reverse (emit-helper exprs
                                                             ()
                                                             (+ 2 indent))))
                               (pad (format ")~%") indent)))
                       ((mul)  (string-append
                               (pad (format "(MUL ~v~%" ptr) indent)
                               (apply string-append
                                      (reverse (emit-helper exprs
                                                            ()
                                                            (+ 2 indent))))
                               (pad (format ")~%") indent)))
                       ((zero) (pad (format "(ZERO ~v)~%" ptr) indent))
                       (else (error "Unknown type " type)))))
          (emit-helper (cdr AST) (cons str acc) indent))))
  
  )