;; BF MIPS Assembly Emitter
;; Jon Simons (simonsj at ccs dot neu dot edu)
;; August, 2007
;; -------------------------------------------------------------------------
;; This module emits MIPS Assembly for the BF IR.
;;
;;   A 30000 byte space is reserved in the .data section for use as
;;   the BF "memory".  A 2 byte scratch space is reserved for string
;;   reading and printing syscalls (which require null-terminated
;;   strings).
;;
;;   Registers are used as follows:
;;
;;     $t0 is the "pointer" to the BF "memory"
;;
;;     $t1 is used for scratch space for INCs and DECs and other stuff
;;
;;     $t2 is used for "mult temps", which are the factors by which INCs
;;         and DECs get multiplied inside of MUL bodies
;;
;;     $t3 is used to store the result of multiplications
;;
;;   Emitted code was tested using PCSpim on Windows, and SPIM on Linux.
;;
;;   This module emits code which is completely oblivious to any integer
;;   wrapping issues.
;; -------------------------------------------------------------------------
(module bf-emit-mips mzscheme
  
  (provide emit)
  
  ;; For AST structs
  (require "../bf-ast-transform.scm")
  ;; Emit Utils
  (require "bf-emit-util.scm")
  
  (require (lib "list.ss"))
  
  ;; emit : (list of EXPRs) -> string
  (define (emit AST)
    (string-append (emit-prologue)
                   (emit-program AST)
                   (emit-epilogue)))
  
  ;; emit-prologue : -> string
  (define (emit-prologue)
    (string-append "        .data~%"
                   "Scratch:~%"
                   "        .space 2~%"
                   "BFmemory:~%"
                   "        .space 30000~%"
                   "~%"
                   "        .text~%"
                   "main:~%"
                   "        la   $t0, BFmemory~%"
                   "~%"))
  
  ;; emit-epilogue : -> string
  (define (emit-epilogue)
    (string-append "        # syscall 10 (exit)~%"
                   "        li   $v0 10~%"
                   "        syscall"))
  
  ;; emit-program : (list of EXPRs) -> string
  (define (emit-program AST)
    (apply string-append (emit-exprs AST #f 8 ())))
  
  ;; Emits a list of strings which correspond to the given
  ;; list of EXPRs, using the given indentation level.
  ;;
  ;; The argument "mult-mode" should be #f for normal emitting,
  ;; and #t for multiplication emitting.
  ;;
  ;; emit-exprs : (list of EXPRs) boolean number (list of strings)
  ;;              -> (list of strings)
  (define (emit-exprs lst mult-mode indent acc)
    (if (null? lst)
        (reverse acc)
        (let* ((e   (car lst))
               (ptr (EXPR-ptr e))
               (N   (EXPR-N   e))
               (str (case (EXPR-type e)
                      ((inc)  (string-append
                               (pad (format "lb   $t1, ~v($t0)~%" ptr)
                                    indent)
                               (if mult-mode
                                   (string-append
                                    (pad (format "mul  $t3, $t2, ~v~%" N)
                                         indent)
                                    (pad (format "add  $t1, $t1, $t3~%")
                                         indent))
                                   (pad (format "addi $t1, ~v~%" N) indent))
                               (pad (format "sb   $t1, ~v($t0)~%" ptr)
                                    indent)))
                      ((dec)  (string-append
                               (pad (format "lb   $t1, ~v($t0)~%" ptr)
                                    indent)
                               (if mult-mode
                                   (string-append
                                    (pad (format "mul  $t3, $t2, ~v~%" N)
                                         indent)
                                    (pad (format "sub  $t1, $t1, $t3~%")
                                         indent))
                                   (pad (format "addi $t1, -~v~%" N)indent))
                               (pad (format "sb   $t1, ~v($t0)~%" ptr)
                                    indent)))
                      ((lt)   (pad (format "addi $t0, -~v~%" N)
                                   indent))
                      ((rt)   (pad (format "addi $t0, ~v~%" N)
                                   indent))
                      ((out)  (string-append
                               (pad (format "lb   $a1, ~v($t0)~%" ptr)
                                    indent)
                               (pad "la   $a0, Scratch~%" indent)
                               (pad "sb   $a1, 0($a0)~%" indent)
                               (pad "li   $v0, 4~%" indent)
                               (pad "syscall # out~%" indent)))
                      ((in)   (string-append
                               (pad (format "la   $a0, Scratch~%") indent)
                               (pad (format "li   $a1, 2~%") indent)
                               (pad (format "li   $v0, 8~%") indent)
                               (pad "syscall # in~%" indent)
                               (pad "lb    $a0, 0($a0)~%" indent)
                               (pad (format "sb   $a0, ~v($t0)~%" ptr)
                                    indent)))
                      ((loop)  (let* ((loop-tmp (gen-loop))
                                      (start    (GENLOOP-start loop-tmp))
                                      (end      (GENLOOP-end   loop-tmp)))
                                 (string-append
                                  (format "~a:~%" start)
                                  (pad (format "lb   $t1, ~v($t0)~%" ptr)
                                       indent)
                                  (pad (format "beqz $t1, ~a~%" end) indent)
                                  (apply string-append
                                         (emit-exprs (EXPR-exprs e)
                                                     #f
                                                     indent
                                                     ()))
                                  (pad (format "b    ~a~%" start) indent)
                                  (format "~a:~%" end))))
                      ((mul)  (let ((skip-label (mul-skip)))
                                (string-append
                                 (pad (format "lb   $t2, ~v($t0)~%" ptr)
                                      indent)
                                 (pad (format "beqz $t2, ~a~%" skip-label)
                                      indent)
                                 (apply string-append
                                        (emit-exprs (EXPR-exprs e)
                                                    #t
                                                    indent
                                                    ()))                                 
                                 (format "~a:~%" skip-label))))
                      ((zero) (pad (format "sb   $zero, ~v($t0)~%" ptr)
                                   indent))
                      (else  (error "unknown token, can't emit")))))
          (emit-exprs (cdr lst) mult-mode indent (cons str acc)))))
  
  )